import re
from datetime import datetime
import alembic.command
import alembic.config
from alembic.script import ScriptDirectory
from tornado.log import app_log

from jupyterhub.orm import Base, Column, Integer, ForeignKey, relationship, JSONDict, Unicode, DateTime, Spawner, Group, User, Boolean
from sqlalchemy.orm import backref
from sqlalchemy import inspect

my_table_names = {'dashboards'}


class Dashboard(Base):
    """"Database class for a Dashboard"""

    __tablename__ = 'dashboards'

    id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(Integer, ForeignKey('users.id', ondelete='CASCADE'), nullable=False)
    user = relationship(User, backref=backref("dashboards_own", uselist=True, cascade='all, delete-orphan'))

    name = Column(Unicode(255))
    description = Column(Unicode(255), default='')

    urlname = Column(Unicode(255), index=True, unique=True, nullable=False)

    created = Column(DateTime, default=datetime.utcnow)

    started = Column(DateTime)

    start_path = Column(Unicode(255), default='') # E.g. which ipynb file should Voila display

    presentation_type = Column(Unicode(255), default='') # Code for framework: voila, streamlit, dash etc

    allow_all = Column(Boolean, index=True, default=True)
    
    # The resulting spawner displaying the finished dashboard, once ready
    final_spawner_id = Column(Integer, ForeignKey('spawners.id', ondelete='SET NULL'))
    final_spawner = relationship(Spawner, foreign_keys=[final_spawner_id], backref=backref('dashboard_final_of', uselist=False))

    group_id = Column(Integer, ForeignKey('groups.id', ondelete='SET NULL'))
    group = relationship(Group, foreign_keys=[group_id], backref=backref('dashboard_visitors_for', uselist=False))

    options = Column(JSONDict)

    # Templates are only used when CDSConfig.spawn_as_viewer=True
    template_parent_id = Column(Integer, ForeignKey('dashboards.id', ondelete='SET NULL'))
    template_children = relationship("Dashboard", backref=backref('template_child_of', uselist=False, remote_side=id))
    

    @property
    def groupname(self):
        if not self.urlname:
            raise Exception('Cannot calculate groupname before urlname is set')
        return 'dash-{}'.format(self.urlname)

    # properties on the dashboard wrapper
    # some APIs get these low-level objects
    # when the dashboard isn't running,
    # for which these should all be False
    active = running = ready = False
    pending = None

    @property
    def orm_dashboard(self):
        return self

    @classmethod
    def find(cls, db, urlname, user=None):
        """Find a Dashboard by urlname.
        Returns None if not found.
        """
        if user is None:
            return db.query(cls).filter(cls.urlname == urlname).first()
        return db.query(cls).filter(cls.urlname == urlname, cls.user_id == user.id).first()

    @classmethod
    def find_template_clone(cls, db, dashboard, viewer_user):
        """Find a Dashboard by urlname.
        Returns None if not found.
        """
        return db.query(cls).filter(cls.template_parent_id == dashboard.id, cls.user_id == viewer_user.id).first()

    def is_orm_user_allowed(self, user):
        if user == self.user:
            return True
        if not self.allow_all:
            if not user in self.group.users:
                return False
        return True

    def __iter__(self):
        """
        Fix for bug in JH 1.1, which was fixed in JH 1.2 by:
        https://github.com/jupyterhub/jupyterhub/commit/84acdd5a7ffb3d79ca04aa86ffcea7e067cce6ee#diff-ea603eff790f5d20b768e093b5340908
        (Actually, the real fix was a little bit earlier, but this is the right place).
        Without this, delete dashboard causes an error when corresponding relationships were expired.
        """
        return iter([self])

    async def clone_for_viewer(self, viewer_user, db):
        existing_dashboard = self.find_template_clone(db, self, viewer_user)
        if existing_dashboard is not None:
            # Clone dashboard already exists, but update and save first if any fields need updating:
            if existing_dashboard.name != self.name \
                    or existing_dashboard.description != self.description \
                    or existing_dashboard.start_path != self.start_path \
                    or existing_dashboard.presentation_type != self.presentation_type \
                    or existing_dashboard.options != self.options \
                    or existing_dashboard.allow_all != False:

                existing_dashboard.name = self.name
                existing_dashboard.description = self.description
                existing_dashboard.start_path = self.start_path
                existing_dashboard.presentation_type = self.presentation_type
                existing_dashboard.options = self.options
                existing_dashboard.allow_all = False

                db.add(existing_dashboard)
                db.commit()

            return existing_dashboard

        # Create a completely new clone dashboard since none exists already
        new_dashboard = Dashboard(
            name=self.name,
            urlname=Dashboard.calc_urlname("{}-{}".format(self.urlname, viewer_user.name), db),
            user=viewer_user.orm_user, 
            description=self.description,
            start_path=self.start_path, 
            presentation_type=self.presentation_type,
            options=self.options,
            allow_all=False,
            template_parent_id=self.id
            )

        db.add(new_dashboard)
        db.commit()
        return new_dashboard

    trailingdash_regex = re.compile(r'\-+$')
    unsafe_regex = re.compile(r'[^a-zA-Z0-9]+')

    @classmethod
    def calc_urlname(cls, dashboard_name, db):
        """
        Generate a unique URL slug based on the given dashboard name.
        """
        base_urlname = re.sub(cls.unsafe_regex, '-', dashboard_name).lower()[:35]

        base_urlname = re.sub(cls.trailingdash_regex, '', base_urlname)

        urlname = base_urlname

        now_unique = False
        counter = 1
        while not now_unique:
            orm_dashboard = Dashboard.find(db=db, urlname=urlname)
            if orm_dashboard is None or counter >= 100:
                now_unique = True
            else:
                urlname = "{}-{}".format(base_urlname, counter)
                counter += 1

        return urlname

class DatabaseSchemaMismatch(Exception):
    """Exception raised when the database schema version does not match

    the current version of JupyterHub.
    """


def check_db_revision(engine):
    """Check the JupyterHub database revision

    After calling this function, an alembic tag is guaranteed to be stored in the db.

    - Checks the alembic tag and raises a ValueError if it's not the current revision
    - If no tag is stored (Bug in Hub prior to 0.8),
      guess revision based on db contents and tag the revision.
    - Empty databases are tagged with the current revision
    """
    # Check database schema version
    current_table_names = set(engine.table_names())

    from .dbutil import _temp_alembic_ini

    with _temp_alembic_ini(engine.url) as ini:
        cfg = alembic.config.Config(ini)
        scripts = ScriptDirectory.from_config(cfg)
        head = scripts.get_heads()[0]
        base = scripts.get_base()

        if not my_table_names.intersection(current_table_names):
            # no tables have been created, stamp with current revision
            app_log.debug("Stamping empty dashboards database with alembic revision %s", head)
            alembic.command.stamp(cfg, head)
            return

        if 'cds_alembic_version' not in current_table_names:
            # Has not been tagged or upgraded before.
            # This should only occur for databases created on cdsdashboards 0.0.11 or earlier

            # Need to identify if this is really an old version or not
            rev = head
            if 'dashboards' in current_table_names:
                inspector = inspect(engine)
                cols = inspector.get_columns('dashboards')
                colnames = [c.get('name','') for c in cols]
                if not 'presentation_type' in colnames:
                    rev = base
                    # presentation_type was added in v0.0.13, so the reason we don't have cds_alembic_version 
                    # is because the db was created before we had db versioning
                # If we DO have dashboards.presentation_type but no cds_alembic_version then this is just because
                # it's a new installation, Dashboards has been created before this first check.
            
            app_log.debug("Stamping dashboards database schema version %s", rev)
            alembic.command.stamp(cfg, rev)

        else:
            alembic_revision = engine.execute(
                'SELECT version_num FROM cds_alembic_version'
            ).first()[0]

            if alembic_revision == base:
                if 'dashboards' in current_table_names:
                    inspector = inspect(engine)
                    cols = inspector.get_columns('dashboards')
                    colnames = [c.get('name','') for c in cols]
                    if 'presentation_type' in colnames:
                        # For people who got stuck in the broken upgrade before - actually they are NOT on base...
                        rev = '260ac5c1a9e0'
                        app_log.debug("Stamping dashboards database schema version %s", rev)
                        alembic.command.stamp(cfg, rev)
                    
    # check database schema version
    # it should always be defined at this point
    alembic_revision = engine.execute(
        'SELECT version_num FROM cds_alembic_version'
    ).first()[0]

    if alembic_revision == head:
        app_log.debug("database dashboards schema version found: %s", alembic_revision)
        pass
    else:
        raise DatabaseSchemaMismatch(
            "Found database schema version {found} != {head}. "
            "Backup your database and run `jupyterhub upgrade-db`"
            " to upgrade to the latest schema.".format(
                found=alembic_revision, head=head
            )
        )



"""
Application for configuring and building the app environments.
"""

import os, re, sys
import logging
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urlparse

from traitlets import Unicode, Integer, Bool, Dict, validate, Any, default, observe, List, TraitError, Union, Callable
from traitlets.config import Application, catch_config_error, SingletonConfigurable
from tornado.httpclient import AsyncHTTPClient
from tornado.httpserver import HTTPServer
import tornado.ioloop
import tornado.options
import tornado.log
from sqlalchemy.exc import OperationalError
from jinja2 import Environment, FileSystemLoader, PrefixLoader, ChoiceLoader
from jupyterhub.services.auth import HubOAuthCallbackHandler
from jupyterhub import __version__ as __jh_version__
from jupyterhub import dbutil
from jupyterhub.traitlets import EntryPointType

from .dashboard import DashboardRepr
from .util import url_path_join
from jupyterhub import orm as jhorm
from .builder.builders import BuildersDict, Builder
from ._data import DATA_FILES_PATH
from .pluggymanager import pm
from . import hookimpl


_TEMPLATE_PATH_OPEN = os.path.join(DATA_FILES_PATH, 'templates-open')
_TEMPLATE_PATH_RESTRICTED = os.path.join(DATA_FILES_PATH, 'templates-restricted')
_TEMPLATE_PATH_COMMON = os.path.join(DATA_FILES_PATH, 'templates-common')


CDS_TEMPLATE_PATHS = [_TEMPLATE_PATH_OPEN, _TEMPLATE_PATH_COMMON]
CDS_TEMPLATE_PATHS_RESTRICTED = [_TEMPLATE_PATH_RESTRICTED, _TEMPLATE_PATH_COMMON]

_all_allowed_presentation_types = ['voila', 'streamlit', 'plotlydash', 'bokeh', 'panel', 'rshiny']


class CDSDashboardsConfig(SingletonConfigurable):

    builder_class = EntryPointType(
        default_value='cdsdashboards.builder.builders.Builder',
        klass=Builder,
        entry_point_group="cdsdashboards.builders",
        help="""The class to use for building dashboard servers.

        Should be a subclass of :class:`cdsdashboards.builder.builders.Builder`.

        May be registered via entry points,
            e.g. `c.cdsdashboards.builders = 'localprocess'`
        """,
    ).tag(config=True)

    server_name_template = Unicode(
        'dash-{urlname}',
        help="""
        How to name the final user server that runs the dashboard. Template vars will be expanded:

        {urlname} : dashboard URL-safe name
        {date} : <current date in YYmmdd format>
        {time} : <current date in HHMMSS format>

        """
    ).tag(config=True)

    presentation_types = List(
        trait=Unicode,
        default_value=_all_allowed_presentation_types,
        minlen=1,
        help="""
        Allowed presentation types for Dashboards. A list, allowed strings are: {}.
        There must be at least one valid entry.
        Default value is all the allowed presentation types.
        Add any custom frameworks to the extra_presentation_types config if you want to augment instead of overwrite the default list.
        """.format(_all_allowed_presentation_types)
    ).tag(config=True)

    extra_presentation_types = List(
        trait=Unicode,
        default_value=[],
        minlen=0,
        help="""
        Extra custom presentation types for Dashboards, to be added to the presentation_types list.
        A list.
        Default value is the empty list.
        """
    ).tag(config=True)

    @property
    def merged_presentation_types(self):
        return self.presentation_types + self.extra_presentation_types

    show_source_git = Bool(
        True,
        help="""
        Allow the user to enter a git repo to fetch files for a dashboard.
        """
    ).tag(config=True)

    default_allow_all = Bool(
        True,
        help="""
        If True (default) then newly-created Dashboards will be accessible to all authenticated JupyterHub users.
        If False, only members of the dashboard's own group will be allowed to access it.
        This flag is passed on to the Dashboard object's allow_all field so can be subsequently overridden in the database.
        """
    ).tag(config=True)

    conda_envs = Union(
        [
            List(trait=Unicode, default_value=[], minlen=0),
            Callable()
        ],
        help="""
        A list of Conda env names for the dashboard creator to select.
        A list.
        Default value is the empty list. Also accepts conda_envs to
        be a callable function which is evaluated for each render of
        the dashboard page.
        """
    ).tag(config=True)

    allow_custom_conda_env = Bool(
        False,
        help="""
        If True then dashboard creators can type any value for the Conda env to use, even if none is provided in the conda_envs list of pre-defined env names.
        If False (default), only Conda env names listed in the conda_envs setting can be selected - if any are present.
        """
    ).tag(config=True)

    spawn_allow_group = Unicode(
        '',
        help="""
        Name of a JupyterHub group whose users should be allowed to spawn servers and create dashboards.
        See also spawn_block_group.
        """
    ).tag(config=True)

    spawn_block_group = Unicode(
        '',
        help="""
        Name of a JupyterHub group whose users should be blocked from spawning servers and creating dashboards.
        If blank, spawn_allow_group will determine which users should be able to spawn.
        If both settings are blank, all users will be allowed to spawn.
        If both are non-blank, spawn_block_group will take priority.
        """
    ).tag(config=True)

    include_auth_state = Bool(
        False,
        help="""
        Whether to include `auth_state` as part of the user information returned by
        dashboards-api/hub-info/user endpoint; default False.

        Warning: `auth_state` can contained sensitive information about the user
        (like credentials).
        """
    ).tag(config=True)

    include_servers = Bool(
        False,
        help="""
        Whether to include the user servers as part of the user information returned by
        dashboards-api/hub-info/user endpoint; default False.
        """
    ).tag(config=True)

    include_servers_state = Bool(
        False,
        help="""
        Whether to include the user servers state as part of the user information returned by
        dashboards-api/hub-info/user endpoint; default False.
        """
    ).tag(config=True)

    spawn_default_options = Bool(
        True,
        help="""
        This applies when the spawner may display options to the user before starting a server.
        If True, a Dashboard will always be started with the default spawner options, so no form will be displayed
        and the dashboard will start without waiting.
        If False (and form options are available in the spawner) then the Dashboard will not start until the
        dashboard creator has accessed the dashboard and provided options to the spawner.
        """
    ).tag(config=True)

    spawn_as_viewer = Bool(
        False,
        help="""
        If True, spawns dashboard as a template owned by the viewing user and accessible only by the viewer.
        Runs multiple dashboard servers per dashboard.
        """
    ).tag(config=True)

    jupyter_startpath_regex = Unicode(
        "",
        help="""
        If not blank, used as a regex that the Start Path of a dashboard must match (if using Jupyter Tree rather than Git Repo as a source).
        This is checked only when a dashboard is created or edited (through the form).
        """
    ).tag(config=True)

class CDSConfigStore():

    _instance = None

    @classmethod
    def get_instance(cls, config):
        """
        Supply a config object to get the singleton CDSDashboardsConfig instance - only normally available from web handlers
        """
        if cls._instance:
            return cls._instance

        cls._instance = CDSDashboardsConfig(config=config)

        return cls._instance


class BuildersStore():

    _instance = None

    @classmethod
    def get_instance(cls, config):
        """
        Supply a config object to get the singleton instance - only normally available from web handlers
        """
        if cls._instance:
            return cls._instance

        cdsconfig = CDSConfigStore.get_instance(config)

        builder_class = cdsconfig.builder_class

        def builder_factory(dashboard):
            return builder_class(dashboard=dashboard, cdsconfig=cdsconfig)

        cls._instance = BuildersDict(builder_factory)
        return cls._instance


cds_tornado_settings = {}



import os.path
import re
from copy import deepcopy
from traitlets import Unicode, Integer, Dict, Bool, validate, default
from traitlets.config import Configurable

from jupyterhub.spawner import _quote_safe
from jupyterhub.traitlets import Command

from ..base import SpawnPermissionsController, CDSConfigStore

def _get_voila_template(args, spawner):

    voila_template = getattr(spawner, 'voila_template', '')

    if voila_template != '':
            args.append('='.join(('{--}template', voila_template)))

    return args

def _get_streamlit_debug(args, spawner):
    try:
        if spawner.debug:
            args.insert(args.index('streamlit')+1, '{--}log_level=debug')
    except ValueError:
        pass
    return args

def _fixed_format(s, **kwargs):
    for k,v in kwargs.items():
        s = s.replace(''.join(('{',k,'}')), v)
    return s

class VariableMixin(Configurable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Merge extra_presentation_launchers config into a copy of builtin_presentation_launchers
        self.merged_presentation_launchers = deepcopy(self.builtin_presentation_launchers)
        for frameworkname, launcher in self.extra_presentation_launchers.items():
            if frameworkname not in self.merged_presentation_launchers:
                self.merged_presentation_launchers[frameworkname] = {}
            for k,v in launcher.items():
                self.merged_presentation_launchers[frameworkname][k] = v

    builtin_presentation_launchers = {
        'voila': {
            #'cmd': ['python3', '-m', 'jhsingle_native_proxy.main'],  # This is the default cmd anyway
            'args': ['--destport=0', 'python3', '{-}m','voila', '{presentation_path}',
                '{--}port={port}',
                '{--}no-browser',
                '{--}Voila.base_url={base_url}/',
                '{--}Voila.server_url=/',
                '{--}Voila.ip=0.0.0.0',
                '{--}Voila.tornado_settings', 'allow_origin={origin_host}',
                '--progressive',
                # If {--}preheat_kernel=False (Volia's default in 0.3.5) then using / as the ready check URL
                # gives a connection refused. /voila/static/ at least gives a 404 so we can tell the 
                # Voila server is available.
                '--ready-check-path=/voila/static/'
                ],
            'extra_args_fn': _get_voila_template
        },
        'streamlit': {
            'args': ['--destport=0', 'streamlit', 'run', '{presentation_path}',
                '{--}server.port={port}',
                '{--}server.headless=True',
                '{--}browser.serverAddress={origin_host}',
                '{--}browser.gatherUsageStats=false'],
            'debug_args': [], # The default is {--}debug, we don't want that
            'extra_args_fn': _get_streamlit_debug # But --log_level=debug has to come earlier in the cmdline
        },
        'plotlydash': {
            'args': ['--destport=0', 'python3', '{-}m','plotlydash_tornado_cmd.main', '{presentation_path}',
                '{--}port={port}'],
            'env': {'DASH_REQUESTS_PATHNAME_PREFIX': '{base_url}/'}
        },
        'bokeh': {
            'args': ['--destport=0', 'python3', '{-}m','bokeh_root_cmd.main', '{presentation_path}',
                '{--}port={port}',
                '{--}allow-websocket-origin={origin_host}',
                '{--}prefix={base_url}',
                '--ready-check-path=/ready-check']
        },
        'panel': {
            'args': ['--destport=0', 'python3', '{-}m','bokeh_root_cmd.main', '{presentation_path}',
                '{--}port={port}',
                '{--}allow-websocket-origin={origin_host}',
                '{--}server=panel',
                '{--}prefix={base_url}',
                '--ready-check-path=/ready-check']
        },
        'rshiny': {
            'args': ['--destport=0', 'python3', '{-}m','rshiny_server_cmd.main', '{presentation_path}',
                '{--}port={port}']
        }

    }

    extra_presentation_launchers = Dict(
        {},
        help="""
        Configuration dict containing details of any custom frameworks that should be made available to Dashboard creators.
        Any new keys added here also need to be added to the c.CDSDashboardsConfig.presentation_types list.
        See cdsdashboards/hubextension/spawners/variablemixin.py in the https://github.com/ideonate/cdsdashboards source code
        for details of the builtin_presentation_launchers dict which shows some examples. This extra_presentation_launchers
        config takes the same format.
        Any keys in extra_presentation_launchersthat also belong to builtin_presentation_launchers will be merged into the
        builtin config, e.g. {'streamlit':{'env':{'STREAMLIT_ENV_VAR':'TEST'}}} will overwrite only the env section of the
        builting streamlit launcher.
        """
    ).tag(config=True)

    default_presentation_cmd = Command(
        ['python3', '-m', 'jhsingle_native_proxy.main'],
        allow_none=False,
        help="""
        The command to run presentations through jhsingle_native_proxy, can be a string or list.
        Default is ['python3', '-m', 'jhsingle_native_proxy.main']
        Change to e.g. ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main'] to ensure start hooks are
        run in the singleuser Docker images.
        """
    ).tag(config=True)

    voila_template = Unicode(
        '',
        help="""
        --template argument to pass to Voila. Default is blank (empty string) to not pass any template to Voila command line.
        """,
    ).tag(config=True)

    proxy_request_timeout = Integer(
        0,
        help="""
        Request timeout in seconds that jhsingle-native-proxy should allow when proxying to the underlying process (e.g. Voila).
        The default of 0 means that no --request-timeout flag will be passed to jhsingle-native-proxy so it will use its own default.
        """,
    ).tag(config=True)

    proxy_ready_timeout = Integer(
        0,
        help="""
        Ready timeout in seconds that jhsingle-native-proxy should allow the underlying process to wait during startup before failing if still
        unable to respond at the --ready-check-path URL.
        The default of 0 means that no --ready-timeout flag will be passed to jhsingle-native-proxy so it will use its own default.
        """,
    ).tag(config=True)

    proxy_force_alive = Bool(
        True,
        help="""
        Whether or not jhsingle-native-proxy should fake activity on its subprocess, always reporting to the hub that activity has happened.
        The default of True means that no flag will be passed to jhsingle-native-proxy so it will use its own default (expected to be --force-alive).
        If False is specified, --no-force-alive will be passed to jhsingle-native-proxy.
        """,
    ).tag(config=True)

    proxy_last_activity_interval = Integer(
        300,
        help="""
        Frequency in seconds that jhsingle-native-proxy should send any recent activity timestamp to the hub.
        If the default of 300 is specified, no --last-activity-interval flag will be passed to jhsingle-native-proxy so it will use its default.
        Otherwise the specified value will be passed to --last-activity-interval.
        """,
    ).tag(config=True)

    proxy_websocket_max_message_size = Integer(
        0,
        help="""
        Max websocket message size allowed by jhsingle-native-proxy, passed as --websocket-max-message-size on the command line.
        The default of 0 means that no --websocket-max-message-size flag will be passed to jhsingle-native-proxy so it will use its own default.
        """,
    ).tag(config=True)

    async def start(self):
        """
        Copy trait values from user_options into the trait attrs of the spawner object
        """

        if self.user_options:
            trait_names = set(self.trait_names()) - {'user_options'}
            for k in trait_names.intersection(self.user_options.keys()):
                merged_trait = self.user_options[k]
                if type(getattr(self, k, None)) == dict:
                    # Merge dicts if one already exists for this trait
                    merged_trait = {**getattr(self, k), **merged_trait}
                setattr(self, k, merged_trait)

        # Any update for cmd needs to be set here (args and env have their own overridden functions)
        presentation_type = self._get_presentation_type()
        if presentation_type != '':
            launcher = self.merged_presentation_launchers[presentation_type]
            if 'cmd' in launcher:
                self.cmd = launcher['cmd']
            else:
                self.cmd = self.default_presentation_cmd

        return await super().start()

    def get_args(self):
        """Return the arguments to be passed after self.cmd

        Doesn't expect shell expansion to happen.

        Also adds self.args at the end in case specified by the config.
        """

        presentation_type = self._get_presentation_type()

        if presentation_type == '':
            return super().get_args()

        launcher = self.merged_presentation_launchers[presentation_type]

        presentation_path = self.user_options.get('presentation_path', '')

        args = []

        # jhsingle-native-proxy --destport $destport --authtype oauth voila `pwd` {--}port={port} {--}no-browser {--}Voila.base_url={base_url}/ {--}Voila.server_url=/ --port $port

        notebook_dir = '.'
        if self.notebook_dir:
            notebook_dir = self.format_string(self.notebook_dir)

        git_repo = self.user_options.get('git_repo', '')
        git_repo_branch = self.user_options.get('git_repo_branch', '')
        repofolder = ''
        if git_repo != '':
            repofolder = self._calc_repo_folder(git_repo)
            args.append('--repo={}'.format(_quote_safe(git_repo)))
            notebook_dir = os.path.join(notebook_dir, repofolder)
            args.append('--repofolder={}'.format(_quote_safe(notebook_dir)))

            if git_repo_branch:
              args.append('--repobranch={}'.format(_quote_safe(git_repo_branch)))

        if presentation_path != '' and not '..' in presentation_path:
            # Should have been validated when dashboard created, but .. is particularly dangerous
            presentation_path = re.sub('^/+', '', presentation_path) # Remove leading slash(es) to ensure it is relative to home folder
            notebook_dir = os.path.join(notebook_dir, presentation_path)

        if 'args' in launcher:
            args.extend(launcher['args'])

        args.append('--presentation-path={}'.format(_quote_safe(notebook_dir)))

        conda_env = self.user_options.get('conda_env', '')
        if conda_env != '':
            args.append('--conda-env=%s' % _quote_safe(conda_env))

        if self.ip:
            args.append('--ip=%s' % _quote_safe(self.ip))

        if self.port:
            args.append('--port=%i' % self.port)

        if self.debug:
            if 'debug_args' in launcher:
                args.extend(launcher['debug_args'])
            else:
                args.append('{--}debug')
            args.append('--debug') # For jhsingle-native-proxy itself

        proxy_request_timeout = getattr(self, 'proxy_request_timeout', 0)
        if proxy_request_timeout:
            args.append('--request-timeout={}'.format(proxy_request_timeout))

        proxy_ready_timeout = getattr(self, 'proxy_ready_timeout', 0)
        if proxy_ready_timeout:
            args.append('--ready-timeout={}'.format(proxy_ready_timeout))

        proxy_force_alive = getattr(self, 'proxy_force_alive', True)
        if proxy_force_alive == False:
            args.append('--no-force-alive')

        proxy_last_activity_interval = getattr(self, 'proxy_last_activity_interval', 300)
        if proxy_last_activity_interval != 300:
            args.append('--last-activity-interval={}'.format(proxy_last_activity_interval))

        proxy_websocket_max_message_size = getattr(self, 'proxy_websocket_max_message_size', 0)
        if proxy_websocket_max_message_size:
            args.append('--websocket-max-message-size={}'.format(proxy_websocket_max_message_size))

        args.extend(self.args)

        if 'extra_args_fn' in launcher and callable(launcher['extra_args_fn']): # Last chance for launcher config to change everything and anything
            args = launcher['extra_args_fn'](args, self)

        return args

    def _get_presentation_type(self):
        """
        Returns the presentation_type (e.g. '' for standard spawner, 'voila', 'streamlit' for named presentation frameworks).
        Throws an exception if the presentation_type doesn't have a launcher configuration in either extra_presentation_launchers
        or builtin_presentation_launchers.
        """
        if self.user_options and 'presentation_type' in self.user_options:
            presentation_type = self.user_options['presentation_type']
            if presentation_type not in self.merged_presentation_launchers:
                raise Exception('presentation type {} has not been registered with the spawner'.format(presentation_type))
            return presentation_type
        return ''

    def get_env(self):
        env = super().get_env()

        presentation_type = self._get_presentation_type()

        if presentation_type != '':

            launcher = self.merged_presentation_launchers[presentation_type]

            if 'env' in launcher:
                presentation_dirname = '.'
                presentation_path = ''
                if self.user_options and 'presentation_path' in self.user_options:
                    presentation_path = self.user_options['presentation_path']
                    presentation_dirname = os.path.dirname(presentation_path)

                self.log.info('presentation_dirname: {}'.format(presentation_dirname))

                for k,v in launcher['env'].items():
                    env[k] = _fixed_format(v,
                        base_url=self.server.base_url,
                        presentation_dirname=presentation_dirname,
                        presentation_path=presentation_path,
                        username=self.user.name
                    )
        return env

    def _calc_repo_folder(self, git_repo):
        s = re.sub('^https?', '', git_repo.lower()) # Remove https and convert to lower case
        s = re.sub('[^a-z0-9]', '-', s) # Replace any non-alphanumeric chars with dash
        s = re.sub('^-+|-+$|-(?=-)', '', s) # Remove dashes from start/end and reduce multiple dashes to just one dash
        return s

    def run_pre_spawn_hook(self):
        if not SpawnPermissionsController.get_instance(CDSConfigStore.get_instance(self.config), self.db).can_user_spawn(self.user.orm_user):
            raise Exception('User {} is not allowed to spawn a server'.format(self.user.name))
        return super().run_pre_spawn_hook()

    def _wrap_options_from_form(outerself, realfn):
        def inner_options_from_form(options):
            """
            If there is an options_form present on a spawner, then when it is submitted by the user,
            it clobbers any existing user_options - which may include 'presentation_*' etc data
            from the dashboard.
            For now, 
            """
            formdata = realfn(options)
            return outerself._postprocess_options_from_form(formdata)

        return inner_options_from_form


    def _postprocess_options_from_form(self, formdata):
        """
        If there is an options_form present on a spawner, then when it is submitted by the user,
        it clobbers any existing user_options - which may include 'presentation_*' etc data
        from the dashboard.
        For now,
        """
        if hasattr(self, 'orm_spawner') and self.orm_spawner and hasattr(self.orm_spawner, 'user_options') \
                    and isinstance(self.orm_spawner.user_options, dict):
            existing_user_options = self.orm_spawner.user_options.copy()
            existing_user_options.update(formdata)
            formdata = existing_user_options
        return formdata

    @validate('options_from_form')
    def _validate_options_from_form(self, proposal):
        """
        Make sure we wrap this custom options_from_form in a function that will preserve dashboard metadata
        """
        return self._wrap_options_from_form(proposal['value'])

    @default("options_from_form")
    def _varmix_options_from_form(self):
        """
        Make sure we wrap the default options_from_form in a function that will preserve dashboard metadata
        May need to override this to something more specific, e.g. in KubeSpawner
        """
        return self._wrap_options_from_form(self._default_options_from_form)  # Spawner class' default


class MetaVariableMixin(type(Configurable)):
    """
    Use this metaclass to ensure VariableMixin occurs earlier in the MRO, so all traits are accessible at the right time.
    """

    def mro(cls):
        mro = super().mro()
        # Take VariableMixin (normally item 4) and put it at item 1
        try:
            vm_index = mro.index(VariableMixin)
            if vm_index > 1 and vm_index < len(mro)-1:
                mro = [mro[0], mro[vm_index]]+ mro[1:vm_index] + mro[vm_index+1:]
        except ValueError:
            pass

        return mro



import asyncio
import concurrent.futures
import inspect


def url_path_join(*pieces):
    """Join components of url into a relative url.

    Use to prevent double slash when joining subpath. This will leave the
    initial and final / in place.

    Copied from `notebook.utils.url_path_join`.
    """
    initial = pieces[0].startswith("/")
    final = pieces[-1].endswith("/")
    stripped = [s.strip("/") for s in pieces]
    result = "/".join(s for s in stripped if s)

    if initial:
        result = "/" + result
    if final:
        result = result + "/"
    if result == "//":
        result = "/"

    return result


class DefaultObjDict(dict):
    def __getattr__(self, name):
        if name in self:
            return self[name]
        return None

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)


def maybe_future(obj):
    """Return an asyncio Future

    Use instead of gen.maybe_future

    For our compatibility, this must accept:

    - asyncio coroutine (gen.maybe_future doesn't work in tornado < 5)
    - tornado coroutine (asyncio.ensure_future doesn't work)
    - scalar (asyncio.ensure_future doesn't work)
    - concurrent.futures.Future (asyncio.ensure_future doesn't work)
    - tornado Future (works both ways)
    - asyncio Future (works both ways)
    """
    if inspect.isawaitable(obj):
        # already awaitable, use ensure_future
        return asyncio.ensure_future(obj)
    elif isinstance(obj, concurrent.futures.Future):
        return asyncio.wrap_future(obj)
    else:
        # could also check for tornado.concurrent.Future
        # but with tornado >= 5 tornado.Future is asyncio.Future
        f = asyncio.Future()
        f.set_result(obj)
        return f


"""add dashboard presentation type

Revision ID: 260ac5c1a9e0
Revises: 2478e08b1ba2
Create Date: 2020-05-27 16:55:34.391858

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '260ac5c1a9e0'
down_revision = '2478e08b1ba2'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('dashboards', sa.Column('presentation_type', sa.Unicode(255), default=''))


def downgrade():
    op.drop_column('dashboards', 'presentation_type')



import sys
from concurrent.futures import ThreadPoolExecutor
import functools

import docker
from docker.utils import kwargs_from_env
from traitlets import Dict, Unicode, Any
from tornado import ioloop
from tornado.log import app_log
from .builders import BuildException
from .processbuilder import ProcessBuilder


class DockerBuilder(ProcessBuilder):

    user = Any()

    @property
    def client(self):
        """single global client instance"""
        cls = self.__class__
        if cls._client is None:
            kwargs = {"version": "auto"}
            if self.tls_config:
                kwargs["tls"] = docker.tls.TLSConfig(**self.tls_config)
            kwargs.update(kwargs_from_env())
            kwargs.update(self.client_kwargs)
            client = docker.APIClient(**kwargs)
            cls._client = client
        return cls._client

    tls_config = Dict(
        config=True,
        help="""Arguments to pass to docker TLS configuration.
        See docker.client.TLSConfig constructor for options.
        """,
    )

    client_kwargs = Dict(
        config=True,
        help="Extra keyword arguments to pass to the docker.Client constructor.",
    )

    _client = None

    def _docker(self, method, *args, **kwargs):
        """wrapper for calling docker methods
        to be passed to ThreadPoolExecutor
        """
        m = getattr(self.client, method)
        return m(*args, **kwargs)

    def docker(self, method, *args, **kwargs):
        """Call a docker method in a background thread
        returns a Future
        """
        fn = functools.partial(self._docker, method, *args, **kwargs)
        return ioloop.IOLoop.current().run_in_executor(self._executor, fn)

    _executor = None

    @property
    def executor(self):
        """single global executor"""
        cls = self.__class__
        if cls._executor is None:
            cls._executor = ThreadPoolExecutor(1)
        return cls._executor

    repo_prefix = Unicode(default_value='cdsuser').tag(config=True)



import sys, re

from tornado.web import authenticated

from jupyterhub.handlers.base import BaseHandler
from jupyterhub.orm import Group, User

from ..orm import Dashboard
from .base import DashboardBaseMixin, check_database_upgrade, spawner_to_dict
from ..util import DefaultObjDict, url_path_join, maybe_future

from .. import hookimpl
from ..pluggymanager import pm
from ..app import BuildersStore, CDSConfigStore


class DashboardBaseHandler(BaseHandler, DashboardBaseMixin):

    def render_template(self, name, sync=False, **ns):
        """
        JupyterHub 1.3+ returns an awaitable from render_template (if new param sync=False).
        Older JH versions just return the template, so this function normalizes everything to an awaitable.
        """
        return maybe_future(super().render_template(name, **ns))

class AllDashboardsHandler(DashboardBaseHandler):

    @authenticated
    @check_database_upgrade
    async def get(self):

        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])

        current_user = await self.get_current_user()

        my_dashboards = current_user.dashboards_own

        if cdsconfig.spawn_as_viewer:
            # Remove clones of template dashboards
            my_dashboards = [d for d in my_dashboards if d.template_parent_id is None]

        visitor_dashboard_groups = self.get_visitor_dashboards(current_user)

        html = await self.render_template(
            "alldashboards.html",
            my_dashboards=my_dashboards,
            visitor_dashboard_groups=visitor_dashboard_groups,
            current_user=current_user,
            can_spawn=self.can_user_spawn(current_user)
        )
        self.write(html)


class BasicDashboardEditHandler(DashboardBaseHandler):

    @authenticated
    @check_database_upgrade
    async def get(self, dashboard_urlname=None):

        current_user = await self.get_current_user()

        if not self.can_user_spawn(current_user):
            return self.send_error(403)

        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])

        dashboard = None
        dashboard_name = self.get_argument('name', '').strip()
        dashboard_description = self.get_argument('description', '').strip()
        dashboard_presentation_type = ''
        dashboard_start_path = self.get_argument('start_path', '').strip()
        dashboard_options = {}
        user_permissions = cdsconfig.default_allow_all and 'anyusers' or 'selectedusers'

        if dashboard_urlname is not None:

            dashboard = Dashboard.find(db=self.db, urlname=dashboard_urlname, user=current_user)

            if dashboard is None:
                return self.send_error(404)

            if current_user.name != dashboard.user.name: # Only allowed to edit your own dashboard
                return self.send_error(403)

            if cdsconfig.spawn_as_viewer and dashboard.template_parent_id is not None:
                html = await self.render_template(
                    "templatedashboard.html",
                    **self.template_vars(dict(
                    dashboard=dashboard,
                    parent_dashboard=dashboard.template_child_of)
                    )
                )
                return self.write(html)

            dashboard_name = dashboard.name
            dashboard_description = dashboard.description
            dashboard_start_path = dashboard.start_path or ''
            dashboard_presentation_type = dashboard.presentation_type
            dashboard_options = dashboard.options
            user_permissions = dashboard.allow_all and 'anyusers' or 'selectedusers'

        # Get List of possible visitor users
        existing_group_users = None
        if dashboard is not None and dashboard.group:
            existing_group_users = dashboard.group.users

        all_users_tuples = self.get_visitor_tuples(current_user.id, existing_group_users)

        source_type = dashboard_options.get('source_type', 'jupytertree')
        git_repo = ''
        git_repo_branch = ''

        if cdsconfig.show_source_git:
            git_repo = dashboard_options.get('git_repo', '')
            git_repo_branch = dashboard_options.get('git_repo_branch', '')
        else:
            source_type = 'jupytertree'

        conda_env = dashboard_options.get('conda_env', '')

        errors = DefaultObjDict()

        merged_presentation_types = cdsconfig.merged_presentation_types

        if callable(cdsconfig.conda_envs):
            all_conda_envs = cdsconfig.conda_envs()
        else:
            all_conda_envs = cdsconfig.conda_envs

        allow_custom_conda_env = cdsconfig.allow_custom_conda_env

        html = await self.render_template(
            "editdashboard.html",
            **self.template_vars(dict(
            dashboard=dashboard,
            dashboard_name=dashboard_name,
            dashboard_description=dashboard_description,
            dashboard_start_path=dashboard_start_path,
            dashboard_presentation_type=dashboard_presentation_type,
            dashboard_options=dashboard_options,
            source_type=source_type,
            git_repo=git_repo,
            git_repo_branch=git_repo_branch,
            conda_env=conda_env,
            presentation_types=merged_presentation_types,
            all_conda_envs=all_conda_envs,
            allow_custom_conda_env=allow_custom_conda_env,
            current_user=current_user,
            user_permissions=user_permissions,
            show_source_git=cdsconfig.show_source_git,
            all_users_tuples=all_users_tuples,
            errors=errors)
            )
        )
        self.write(html)

    def template_vars(self, d):
        return d

    @authenticated
    @check_database_upgrade
    async def post(self, dashboard_urlname=None):

        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])

        current_user = await self.get_current_user()

        if not self.can_user_spawn(current_user):
            return self.send_error(403)

        dashboard = None
        group = None

        if dashboard_urlname is not None:
            # Edit (not new)

            dashboard = Dashboard.find(db=self.db, urlname=dashboard_urlname, user=current_user)

            if dashboard is None:
                return self.send_error(404)

            if current_user.name != dashboard.user.name:
                return self.send_error(403)

            if cdsconfig.spawn_as_viewer and dashboard.template_parent_id is not None:
                return self.send_error(403)

            group = dashboard.group

        dashboard_name = self.get_argument('name').strip()

        dashboard_description = self.get_argument('description').strip()

        dashboard_presentation_type = self.get_argument('presentation_type').strip()

        dashboard_start_path = self.get_argument('start_path').strip()

        errors = DefaultObjDict()

        # source type (jupytertree or gitrepo) - to be added to dashboard_options later
        git_repo = ''
        git_repo_branch = ''
        source_type = self.get_argument('source_type', '').strip()

        if cdsconfig.show_source_git and source_type == 'gitrepo':
            git_repo = self.get_argument('git_repo', '').strip()
            git_repo_branch = self.get_argument('git_repo_branch', '').strip()

            if git_repo != '':
                if not re.match('^((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(/)?$', git_repo):
                    errors.git_repo = 'Please enter a valid git repo URL'
        else:
            source_type = 'jupytertree'

        # Presentation basics

        if dashboard_name == '':
            errors.name = 'Please enter a name'
        elif not self.name_regex.match(dashboard_name):
            errors.name = 'Please use letters and digits (start with one of these), and then spaces or these characters _-!@$()*+?<>\'". Max 100 chars.'

        jupyter_startpath_regex = None
        if source_type == 'jupytertree' and cdsconfig.jupyter_startpath_regex != "":
            try:
                jupyter_startpath_regex = re.compile(cdsconfig.jupyter_startpath_regex)
            except re.error as rerr:
                self.log.warn('jupyter_startpath_regex ({}) cannot compile: {}'.format(cdsconfig.jupyter_startpath_regex, rerr))

        if '..' in dashboard_start_path:
            errors.start_path = 'Path must not contain ..'
        elif len(dashboard_start_path) and dashboard_start_path[0] == '/':
            errors.start_path = 'Path must be relative to Jupyter tree home or Git repo root (not starting with /)'
        elif not self.start_path_regex.match(dashboard_start_path):
            errors.start_path = 'Please enter valid URL path characters'
        elif jupyter_startpath_regex is not None and not jupyter_startpath_regex.match(dashboard_start_path):
            errors.start_path = 'Path does not match the pattern specified by your admin ({})'.format(cdsconfig.jupyter_startpath_regex)

        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])

        merged_presentation_types = cdsconfig.merged_presentation_types
        all_conda_envs = cdsconfig.conda_envs
        allow_custom_conda_env = cdsconfig.allow_custom_conda_env

        if not dashboard_presentation_type in merged_presentation_types:
            errors.presentation_type = 'Framework {} invalid - it must be one of the allowed types: {}'.format(
                dashboard_presentation_type, ', '.join(merged_presentation_types)
                )

        # Visitors allowed
        selected_users = self.get_arguments('selected_users[]')

        selected_users_orm = self.db.query(User).filter(User.name.in_(selected_users)).all()

        user_permissions = self.get_argument('user_permissions', cdsconfig.default_allow_all and 'anyusers' or 'selectedusers').strip()

        if user_permissions not in ('anyusers', 'selectedusers'):
            user_permissions = cdsconfig.default_allow_all and 'anyusers' or 'selectedusers'
            errors.user_permissions = 'You must choose either anyusers or selectedusers'

        # Dashboard options
        dashboard_options = {}

        conda_env = self.get_argument('conda_env', '').strip()

        if conda_env != '':
            if allow_custom_conda_env:
                if not self.conda_env_regex.match(conda_env):
                    errors.conda_env = 'Please use letters and digits, spaces or these characters _-!@$()*+?<>/\\\'".'
                    conda_env = ''
            elif conda_env not in all_conda_envs:
                errors.conda_env = 'Please select a valid Conda env (\'{}\' is not in the allowed list)'.format(conda_env)
                conda_env = ''

        dashboard_options['source_type'] = source_type
        dashboard_options['git_repo'] = git_repo
        dashboard_options['git_repo_branch'] = git_repo_branch
        dashboard_options['conda_env'] = conda_env


        if len(errors) == 0:
            db = self.db

            try:

                if dashboard is None:

                    urlname = Dashboard.calc_urlname(dashboard_name, self.db)

                    self.log.debug('Final urlname is '+urlname)

                    dashboard = Dashboard(
                        name=dashboard_name, urlname=urlname, user=current_user.orm_user,
                        description=dashboard_description, start_path=dashboard_start_path,
                        presentation_type=dashboard_presentation_type,
                        options=dashboard_options,
                        allow_all=user_permissions=='anyusers'
                        )
                    self.log.debug('dashboard urlname '+dashboard.urlname+', main name '+dashboard.name)

                else:
                    dashboard.name = dashboard_name
                    dashboard.description = dashboard_description
                    dashboard.start_path = dashboard_start_path
                    dashboard.presentation_type = dashboard_presentation_type
                    dashboard.options = dashboard_options
                    allow_all=user_permissions=='anyusers'
                    dashboard.allow_all=allow_all

                if group is None:
                    group = Group.find(db, dashboard.groupname)
                    if group is None:
                        group = Group(name=dashboard.groupname)
                        self.db.add(group)
                    dashboard.group = group

                db.add(dashboard)

                if self.sync_group(group, selected_users_orm):
                    db.add(group)

                db.commit()

                # Now cancel any existing build and force a rebuild

                builders_store = BuildersStore.get_instance(self.settings['config'])
                builder = builders_store[dashboard]

                dashboard_user = self._user_from_orm(dashboard.user.name)

                next_page = await self.maybe_start_build(dashboard, dashboard_user, True)
                # next_page can be progress/options/dashboard, but actually can't be anything because with force_start==True
                # it will have set things off async!

            except Exception as e:
                errors.all = str(e)

        if len(errors):

            git_repo = dashboard_options['git_repo'] = dashboard_options.get('git_repo', '')
            git_repo_branch = dashboard_options['git_repo_branch'] = dashboard_options.get('git_repo_branch', '')
            conda_env = dashboard_options['conda_env'] = dashboard_options.get('conda_env', '')
            all_users_tuples = self.get_visitor_tuples(current_user.id, selected_users_orm)

            html = await self.render_template(
                "editdashboard.html",
                **self.template_vars(dict(
                dashboard=dashboard,
                dashboard_name=dashboard_name,
                dashboard_description=dashboard_description,
                dashboard_start_path=dashboard_start_path,
                dashboard_presentation_type=dashboard_presentation_type,
                dashboard_options=dashboard_options,
                user_permissions=user_permissions,
                git_repo=git_repo,
                git_repo_branch=git_repo_branch,
                source_type=source_type,
                conda_env=conda_env,
                presentation_types=merged_presentation_types,
                all_conda_envs=all_conda_envs,
                allow_custom_conda_env=allow_custom_conda_env,
                show_source_git=cdsconfig.show_source_git,
                all_users_tuples=all_users_tuples,
                errors=errors,
                current_user=current_user))
            )
            return self.write(html)

        self.redirect(url_path_join(self.settings['base_url'], "hub", "dashboards", dashboard.urlname))


class DashboardOptionsHandler(DashboardBaseHandler):

    @authenticated
    @check_database_upgrade
    async def get(self, dashboard_urlname):

        current_user = await self.get_current_user()

        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])

        dashboard = Dashboard.find(db=self.db, urlname=dashboard_urlname)

        if dashboard is None:
            return self.send_error(404)

        if not dashboard.is_orm_user_allowed(current_user.orm_user):
            return self.send_error(403)

        builders_store = BuildersStore.get_instance(self.settings['config'])
        builder = builders_store[dashboard]

        dashboard_user = self._user_from_orm(dashboard.user.name)

        next_page = await self.maybe_start_build(dashboard, dashboard_user, False)

        if next_page == 'options':
            self.log.info('Display options form')
            options_not_allowed = False
            need_user_options_form = builder._needs_user_options

            if current_user.name != dashboard.user.name:
                need_user_options_form = None
                options_not_allowed = True

            html = await self.render_template(
                "dashboardoptions.html",
                dashboard=dashboard,
                current_user=current_user,
                dashboard_user=dashboard.user,
                user_options_form=need_user_options_form,
                options_not_allowed=options_not_allowed
            )
            return self.write(html)

        # The dashboard spawn is in progress, go to the main build page
        return self.redirect(url_path_join(self.settings['base_url'], "hub", "dashboards", dashboard_urlname))

    @authenticated
    @check_database_upgrade
    async def post(self, dashboard_urlname):

        current_user = await self.get_current_user()

        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])

        dashboard = Dashboard.find(db=self.db, urlname=dashboard_urlname)

        if dashboard is None:
            return self.send_error(404)

        if not dashboard.is_orm_user_allowed(current_user.orm_user):
            return self.send_error(403)

        if current_user.name != dashboard.user.name:
            return self.send_error(403)

        form_options = {}
        for key, byte_list in self.request.body_arguments.items():
            form_options[key] = [bs.decode('utf8') for bs in byte_list]
        for key, byte_list in self.request.files.items():
            form_options["%s_file" % key] = byte_list


        builders_store = BuildersStore.get_instance(self.settings['config'])
        builder = builders_store[dashboard]

        next_page = await self.maybe_start_build(dashboard, current_user, True, form_options)

        return self.redirect(url_path_join(self.settings['base_url'], "hub", "dashboards", dashboard_urlname))


class MainViewDashboardHandler(DashboardBaseHandler):

    @authenticated
    @check_database_upgrade
    async def get(self, dashboard_urlname=''):

        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])

        current_user = await self.get_current_user()

        dashboard = self.db.query(Dashboard).filter(Dashboard.urlname==dashboard_urlname).one_or_none()

        if dashboard is None:
            return self.send_error(404)

        if not dashboard.is_orm_user_allowed(current_user.orm_user):
            return self.send_error(403)

        dashboard_user = self._user_from_orm(dashboard.user.name)

        if cdsconfig.spawn_as_viewer and dashboard_user != current_user:
            # Clone the dashboard as a template
            new_dashboard = await dashboard.clone_for_viewer(current_user, self.db)
            return self.redirect(url_path_join(self.settings['base_url'], "hub", "dashboards", new_dashboard.urlname))

        next_page = await self.maybe_start_build(dashboard, dashboard_user)

        if next_page == 'options':
            self.log.debug('Redirect to dashboard/options')
            return self.redirect(url_path_join(self.settings['base_url'], "hub", "dashboards", dashboard_urlname, 'options'))

        if next_page == 'dashboard':
            return self.redirect(url_path_join(self.settings['base_url'], "user", dashboard_user.name, dashboard.final_spawner.name))

        base_url = self.settings['base_url']

        html = await self.render_template(
            "viewdashboard.html",
            dashboard=dashboard,
            current_user=current_user,
            dashboard_user=dashboard_user,
            progress_url=url_path_join(base_url, 'hub', 'dashboards-api', dashboard_urlname, 'progress')
        )
        self.write(html)


class ClearErrorDashboardHandler(DashboardBaseHandler):

    @authenticated
    @check_database_upgrade
    async def get(self, dashboard_urlname=''):

        current_user = await self.get_current_user()

        dashboard = self.db.query(Dashboard).filter(Dashboard.urlname==dashboard_urlname).one_or_none()

        if dashboard is None:
            pass # Redirect anyway, to let regular MainViewDashboardHandler handle the error

        elif dashboard.is_orm_user_allowed(current_user.orm_user):
            builders_store = BuildersStore.get_instance(self.settings['config'])

            builder = builders_store[dashboard]

            if not builder.pending and builder._build_future and builder._build_future.done() and builder._build_future.exception():
                builder._build_future = None

        self.redirect(url_path_join(self.settings['base_url'], "hub", "dashboards", dashboard_urlname))


class UpgradeDashboardsHandler(DashboardBaseHandler):

    @authenticated
    async def get(self):

        from ..dbutil import is_upgrade_needed

        engine = self.db.get_bind()
        if not is_upgrade_needed(engine):
            return self.redirect("{}".format(self.settings['base_url']))

        current_user = await self.get_current_user()

        is_sqlite = str(engine.url).startswith('sqlite:///')

        html = await self.render_template(
            "upgrade-db.html",
            is_admin=current_user.admin,
            error='',
            is_sqlite=is_sqlite,
            success=False
        )
        self.write(html)

    @authenticated
    async def post(self):

        current_user = await self.get_current_user()

        if not current_user.admin:
            return self.send_error(403)

        error = ''
        success = True

        from ..dbutil import upgrade_if_needed

        engine = self.db.get_bind()

        is_sqlite = str(engine.url).startswith('sqlite:///')

        try:
            upgrade_if_needed(engine, log=self.log)
        except Exception as e:
            success = False
            error = str(e)

        html = await self.render_template(
            "upgrade-db.html",
            is_admin=current_user.admin,
            error=error,
            success=success,
            is_sqlite=is_sqlite
        )
        self.write(html)


class GroupsAllHandler(DashboardBaseHandler):

    @authenticated
    async def get(self):

        current_user = await self.get_current_user()

        if not current_user.admin:
            return self.send_error(403)

        groups = self.db.query(Group).order_by(Group.name).all()

        html = await self.render_template(
            "allgroups.html",
            groups=groups
        )
        self.write(html)

class GroupsSingleHandler(DashboardBaseHandler):

    @authenticated
    async def get(self, groupname):

        current_user = await self.get_current_user()

        if not current_user.admin:
            return self.send_error(403)

        group = Group.find(self.db, groupname)

        if group is None:
            return self.send_error(404)

        errors = DefaultObjDict()

        existing_group_users = group.users

        all_users_tuples = self.get_visitor_tuples(None, existing_group_users)

        html = await self.render_template(
            "editgroup.html",
            group=group,
            errors=errors,
            all_users_tuples=all_users_tuples
        )
        self.write(html)

    @authenticated
    async def post(self, groupname):

        current_user = await self.get_current_user()

        if not current_user.admin:
            return self.send_error(403)

        group = Group.find(self.db, groupname)

        if group is None:
            return self.send_error(404)

        errors = DefaultObjDict()

        existing_group_users = group.users

        selected_users = self.get_arguments('selected_users[]')

        selected_users_orm = self.db.query(User).filter(User.name.in_(selected_users)).all()

        if self.sync_group(group, selected_users_orm):
            self.db.add(group)
            self.db.commit()

        self.redirect('{}hub/groupslist'.format(self.settings['base_url']))


# Register plugin hooks so we use the Basic handlers by default, unless overridden

@hookimpl
def get_hubextension_main_DashboardEditHandler():
    return BasicDashboardEditHandler

pm.register(sys.modules[__name__])

DashboardEditHandler = pm.hook.get_hubextension_main_DashboardEditHandler()[0]



import os
import json

from traitlets.config import LoggingConfigurable
from traitlets import Integer, Unicode, Bool
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError
from tornado.log import app_log
from tornado import web, gen


class DashboardRepr(LoggingConfigurable):
    """Object for an individual app/server for a user, such as files or repo and allowed visitors"""

    hub_api_token = Unicode(help="The API token for the Hub")
    hub_url = Unicode(help="The URL of the Hub")

    retries = Integer(
        4,
        config=True,
        help="""Number of attempts to make on Hub API requests.

        Adds resiliency to intermittent Hub failures,
        most commonly due to Hub, proxy, or ingress interruptions.
        """
    )
    retry_delay = Integer(
        4,
        config=True,
        help="""
        Time (seconds) to wait between retries for Hub API requests.

        Time is scaled exponentially by the retry attempt (i.e. 2, 4, 8, 16 seconds)
        """
    )

    async def get_user_data(self, username):
        resp = await self.api_request(
            'users/%s' % username,
            method='GET',
        )
        body = json.loads(resp.body.decode('utf-8'))
        return body

    async def get_app_server(self, username, server_name):
        user_data = await self.get_user_data(username)

        self.log.debug(user_data)

        if not server_name:
            return user_data.server
            
        if 'servers' in user_data and server_name in user_data['servers']:
            return user_data['servers'][server_name]
        
        return None

    async def api_request(self, url, *args, **kwargs):
        """Make an API request to JupyterHub"""
        headers = kwargs.setdefault('headers', {})
        headers.update({'Authorization': 'token %s' % self.hub_api_token})
        hub_api_url = os.getenv('JUPYTERHUB_API_URL', '') or self.hub_url + 'hub/api/'
        request_url = hub_api_url + url

        self.log.info("API call to {} with token {}".format(request_url, self.hub_api_token))

        req = HTTPRequest(request_url, *args, **kwargs, validate_cert=False)
        retry_delay = self.retry_delay
        for i in range(1, self.retries + 1):
            try:
                return await AsyncHTTPClient().fetch(req)
            except HTTPError as e:
                # swallow 409 errors on retry only (not first attempt)
                if i > 1 and e.code == 409 and e.response:
                    self.log.warning("Treating 409 conflict on retry as success")
                    return e.response
                # retry requests that fail with error codes greater than 500
                # because they are likely intermittent issues in the cluster
                # e.g. 502,504 due to ingress issues or Hub relocating,
                # 599 due to connection issues such as Hub restarting
                if e.code >= 500:
                    self.log.error("Error accessing Hub API (using %s): %s", request_url, e)
                    if i == self.retries:
                        # last api request failed, raise the exception
                        raise
                    await gen.sleep(retry_delay)
                    # exponential backoff for consecutive failures
                    retry_delay *= 2
                else:
                    raise



import json
import os
import getpass
import shutil
import sys

from sudospawner import SudoSpawner
from traitlets import Unicode

from .variablemixin import VariableMixin, MetaVariableMixin
from sudospawner.mediator import spawn, kill
from tornado.options import parse_command_line
from tornado import log
app_log = log.app_log

class VariableSudoSpawner(SudoSpawner, VariableMixin, metaclass=MetaVariableMixin):

    async def start(self):
        """
        Copy trait values from user_options into the trait attrs of the spawner object
        """

        if self.user_options:
            trait_names = set(self.trait_names()) - {'user_options'}
            for k in trait_names.intersection(self.user_options.keys()):
                merged_trait = self.user_options[k]
                if type(getattr(self, k, None)) == dict:
                    # Merge dicts if one already exists for this trait
                    merged_trait = {**getattr(self, k), **merged_trait}
                setattr(self, k, merged_trait)

        # Any update for cmd needs to be set here (args and env have their own overridden functions)
        presentation_type = self._get_presentation_type()
        if presentation_type != '':
            self.sudospawner_path = shutil.which('cds_sudospawner') or 'cds_sudospawner'
            launcher = self.merged_presentation_launchers[presentation_type]
            if 'cmd' in launcher:
                self.cmd = launcher['cmd']
            else:
                self.cmd = self.default_presentation_cmd

        return await super().start()


def mediator(*args, **kwargs):
    parse_command_line()
    app_log.debug("Starting mediator for %s", getpass.getuser())
    try:
        kwargs = json.load(sys.stdin)
    except ValueError as e:
        app_log.error("Expected JSON on stdin, got %s" % e)
        sys.exit(1)

    action = kwargs.pop('action')
    if action == 'kill':
        kill(**kwargs)
    elif action == 'spawn':
        spawn("jhsingle-native-proxy", **kwargs)
    else:
        raise TypeError("action must be 'spawn' or 'kill'")


from tornado.log import app_log

from .builders import Builder, BuildException
from ..util import maybe_future, url_path_join

class ProcessBuilder(Builder):
 
    async def start(self, dashboard, dashboard_user, form_options=None, spawn_default_options=True):
        """Start the dashboard

        Returns:
          (str, str, None): the (new_server_name, new_server_options, None) of the new dashboard server.
          or
          (None, None, str): (None, None, user_options_form) if a user options form needs to be presented to the user.

        """

        app_log.info('Starting Builder start function')

        self.event_queue = []

        self.add_progress_event({'progress': 10, 'message': 'Starting builder'})

        self._build_pending = True

        ns = self.template_namespace()

        new_server_name = self.format_string(self.cdsconfig.server_name_template, ns=ns)

        new_server_options = {}

        if not self.allow_named_servers:
            raise BuildException(400, "Named servers are not enabled.")

        spawner = dashboard_user.spawners[new_server_name] # Could be orm_spawner or Spawner wrapper

        if spawner.ready:
            # include notify, so that a server that died is noticed immediately
            # set _spawn_pending flag to prevent races while we wait
            spawner._spawn_pending = True
            try:
                state = await spawner.poll_and_notify()
            finally:
                spawner._spawn_pending = False

        # Does this spawner need user options?
        if not spawn_default_options:

            if not form_options:
                
                spawner_options_form = await spawner.get_options_form()
                if spawner_options_form:
                    app_log.info('Options form is present')
                    return (None, None, spawner_options_form) # Tell caller that we need to go to the options form

            else:

                try:
                    user_options = await maybe_future(spawner.options_from_form(form_options))
                    new_server_options.update(user_options)
                except Exception as e:
                    app_log.error(
                        "Failed to spawn dashboard server with form", exc_info=True
                    )
                    spawner_options_form = await spawner.get_options_form()
                    if spawner_options_form:
                        app_log.info('Options form is present after failure')
                        return (None, None, spawner_options_form)
                        
        # Dashboard-specific options
        git_repo = dashboard.options.get('git_repo', '')
        git_repo_branch = dashboard.options.get('git_repo_branch', '')
        conda_env = dashboard.options.get('conda_env', '')

        new_server_options.update({
            'presentation_type': dashboard.presentation_type or 'voila',
            'presentation_path': dashboard.start_path,
            'git_repo': git_repo,
            'git_repo_branch': git_repo_branch,
            'conda_env': conda_env})
        
        if 'environment' not in new_server_options:
            new_server_options['environment'] = {}

        if not self.cdsconfig.spawn_as_viewer:
            new_server_options['environment'].update({
                    'JUPYTERHUB_ANYONE': '{}'.format(dashboard.allow_all and '1' or '0'),
                    'JUPYTERHUB_GROUP': '{}'.format(dashboard.groupname)
                    })

        return (new_server_name, new_server_options, None)



import re
from datetime import datetime
from collections import defaultdict
from asyncio import sleep, CancelledError

from async_generator import aclosing
from tornado.web import HTTPError
from tornado.log import app_log

from jupyterhub.orm import User, Group
from jupyterhub.utils import iterate_until

from ..util import maybe_future
from ..orm import Dashboard
from ..app import BuildersStore, CDSConfigStore
from ..dbutil import is_upgrade_needed
from ..util import DefaultObjDict, url_path_join

def spawner_to_dict(spawner):
    name = spawner.name
    active = spawner.active
    id = 'named-{}'.format(name)
    
    if name == '':
        id = 'default'
        name = 'Default Server'
        
    return DefaultObjDict({'name': name, 'active': active, 'id': id, 'orm_spawner': spawner.orm_spawner})

def check_database_upgrade(f):
    def handler(self, *args, **kwargs):
        engine = self.db.get_bind()
        if is_upgrade_needed(engine):
            return self.redirect("{}hub/dashboards-db-upgrade".format(self.settings['base_url']))
        return f(self, *args, **kwargs)
    return handler


class SpawnPermissionsController():

    _instance = None

    @classmethod
    def get_instance(cls, cdsconfig, db):
        """
        Supply a config object to get the singleton instance - only normally available from web handlers
        """
        if cls._instance:
            return cls._instance

        if cdsconfig is None:
            raise Exception('SpawnPermissionsController must be run first within a handler - please visit any JupyterHub webpage then try again')
        
        cls._instance = cls(cdsconfig, db)
        return cls._instance

    def __init__(self, cdsconfig, db):
        self.spawn_allow_group = cdsconfig.spawn_allow_group
        self.spawn_block_group = cdsconfig.spawn_block_group

        self.spawn_allow_group_orm = None
        self.spawn_block_group_orm = None

        created_group = False

        if self.spawn_allow_group != '':
            group = Group.find(db, self.spawn_allow_group)
            if group is None:
                group = Group(name=self.spawn_allow_group)
                db.add(group)
                created_group = True
            self.spawn_allow_group_orm = group

        if self.spawn_block_group != '':
            group = Group.find(db, self.spawn_block_group)
            if group is None:
                group = Group(name=self.spawn_block_group)
                db.add(group)
                created_group = True
            self.spawn_block_group_orm = group

        if created_group:
            self.db.commit()

    def can_user_spawn(self, orm_user):
        if self.spawn_block_group_orm:
            if orm_user in self.spawn_block_group_orm.users:
                return False
        elif self.spawn_allow_group_orm:
            return orm_user in self.spawn_allow_group_orm.users
        return True


class DashboardBaseMixin:

    name_regex = re.compile(r'^[a-zA-Z0-9][a-zA-Z0-9\._\- \/~\:\!\@\$#\[\]\(\)\&\*\+\?\<\>\.\'\" ,;\%\=]{0,99}$')

    conda_env_regex = re.compile(r'^[a-zA-Z0-9\._\- \/~\:\!\@\$#\[\]\(\)\&\*\+\?\<\>\.\'\" ,;\%\=]+$')

    start_path_regex = re.compile(r'^[A-Za-z0-9\-\._~\:\/\?#\[\]@\!\$\&\(\)\*\+\'\" ,;\%\=]*$')

    @staticmethod
    async def pipe_spawner_progress(dashboard_user, new_server_name, builder):
        
        while True:

            await sleep(0.01)

            if builder._build_future and builder._build_future.done():
                break

            if new_server_name in dashboard_user.spawners and dashboard_user.spawners[new_server_name].pending \
                and dashboard_user.spawners[new_server_name]._spawn_future:

                spawner = dashboard_user.spawners[new_server_name]

                app_log.debug('Found spawner for progress')

                async with aclosing(
                    iterate_until(maybe_future(builder._build_future), spawner._generate_progress())
                ) as spawnevents:
                    try:
                        async for event in spawnevents:
                            if 'message' in event:
                                builder.add_progress_event({
                                    'progress': 95, 'message': 'Spawner progress: {}'.format(event['message'])
                                    })
                    except CancelledError:
                        pass

                break 

    def get_visitor_dashboards(self, user):
        orm_dashboards = set()

        for group in user.groups:
            if group.dashboard_visitors_for:
                orm_dashboards.add(group.dashboard_visitors_for)

        orm_dashboards.update(self.db.query(Dashboard).filter(Dashboard.allow_all==True).filter(Dashboard.user != user.orm_user).all())

        user_dashboard_groups = defaultdict(list)

        for dash in orm_dashboards:
            if dash.user: # Just in case it's null due to db corruption
                user_dashboard_groups[dash.user.name].append(dash)

        return user_dashboard_groups

    def get_visitor_users(self, exclude_user_id=None):
        if exclude_user_id is not None:
            return self.db.query(User).filter(User.id != exclude_user_id).all()
        return self.db.query(User).all()

    def get_visitor_tuples(self, exclude_user_id=None, existing_group_users=None):
        possible_visitor_users = self.get_visitor_users(exclude_user_id)

        visitor_users = []

        if existing_group_users is not None:
            visitor_users = existing_group_users

        all_tuples = [(True, user.name) for user in visitor_users] + [(False, user.name) for user in set(possible_visitor_users) - set(visitor_users)]
        all_tuples.sort(key=lambda x: x[1])
        return all_tuples

    def sync_group(self, group, visitor_users):
        """
        Make sure all allowed JupyterHub users are part of this group
        Returns True if changes made, False otherwise
        """
        unwantedusers = set(group.users) - set(visitor_users)
        newusers = set(visitor_users) - set(group.users)

        if len(unwantedusers) + len(newusers) > 0:

            for user in unwantedusers:
                group.users.remove(user)

            for user in newusers:
                group.users.append(user)

            return True

        return False

    async def maybe_start_build(self, dashboard, dashboard_user, force_start=False, user_options=None):

        builders_store = BuildersStore.get_instance(self.settings['config'])

        builder = builders_store[dashboard]

        user = self._user_from_orm(dashboard_user)

        # If dashboard is already completely running, just redirect there.

        if not force_start:

            if builder.pending or builder._build_future is not None:
                return 'progress' # already mid-build, or exception to show
            
            if builder._needs_user_options:
                return 'options'

            if dashboard.final_spawner is not None:
                final_spawner = user.spawners[dashboard.final_spawner.name]

                if final_spawner.ready:
                    return 'dashboard'
                #return 'progress' # Need to attach to progress events to await spawn (or stop)

            # Assume builder.pending == False and builder._build_future is None

                

        # There should only really be a build in progress if we have force_start == True


        self.log.debug('starting builder')

        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])
        spawn_default_options = cdsconfig.spawn_default_options


        def do_final_build(f):
            self.log.debug('In do_final_build')
            if f.cancelled() or f.exception() is None:
                builder._build_future = None
            builder._build_pending = False

        existing_build_future = None

        if builder.pending and builder._build_future and not builder._build_future.done():
            if force_start:
                builder._build_future.cancel()
            existing_build_future = builder._build_future

        async def do_build():

            if existing_build_future:
                await existing_build_future

            # Delete existing final spawner if it exists
            await self.maybe_delete_existing_server(dashboard.final_spawner, dashboard_user)

            (new_server_name, new_server_options, need_user_options_form) =  await builder.start(dashboard, dashboard_user, user_options, spawn_default_options)

            if need_user_options_form:
                builder.add_progress_event({'progress': 80, 'message': 'Needs user options form'})
                builder._needs_user_options = need_user_options_form
                return
            
            self.log.debug('Confirming user options not needed')
            builder._needs_user_options = False

            builder.add_progress_event({'progress': 80, 'message': 'Starting up final server for Dashboard, after build'})

            maybe_future(self.pipe_spawner_progress(dashboard_user, new_server_name, builder))

            await self.spawn_single_user(dashboard_user, new_server_name, options=new_server_options)

            self.db.expire(dashboard) # May have changed during async code above

            final_spawner = None

            if new_server_name in dashboard_user.orm_user.orm_spawners:
                final_spawner = dashboard_user.spawners[new_server_name]
                dashboard.final_spawner = dashboard_user.orm_user.orm_spawners[new_server_name]

            # TODO if not, then what?

            dashboard.started = datetime.utcnow()

            self.db.commit()

            if final_spawner and not final_spawner.ready:
                await maybe_future(getattr(final_spawner, '_{}_future'.format(final_spawner.pending), None))

        builder._build_pending = True
        builder._build_future = maybe_future(do_build())

        builder._build_future.add_done_callback(do_final_build)

        return 'progress'

    async def maybe_delete_existing_server(self, orm_spawner, dashboard_user):
        if not orm_spawner:
            return False
        
        server_name = orm_spawner.name

        user = self._user_from_orm(dashboard_user)

        spawner = user.spawners[server_name]

        if server_name:
            if not self.allow_named_servers:
                raise HTTPError(400, "Named servers are not enabled.")
            if server_name not in user.orm_spawners:
                self.log.debug("%s does not exist anyway", spawner._log_name)
                return False
        else:
            raise HTTPError(400, "Cannot delete the default server")

        if spawner.pending == 'stop':
            self.log.debug("%s already stopping", spawner._log_name)

            if spawner._stop_future:
                await spawner._stop_future

        if spawner.ready or spawner.pending == 'spawn':

            if spawner.pending == 'spawn':
                self.log.debug("%s awaiting spawn to finish so can shut it down", spawner._log_name)
                await spawner._spawn_future

            # include notify, so that a server that died is noticed immediately
            status = await spawner.poll_and_notify()
            if status is None:
                stop_future = await self.stop_single_user(user, server_name)
                await stop_future

        self.log.info("Deleting spawner %s", spawner._log_name)
        self.db.delete(spawner.orm_spawner)
        user.spawners.pop(server_name, None)
        self.db.commit()

    def can_user_spawn(self, user):
        if user is None:
            return False
        cdsconfig = CDSConfigStore.get_instance(self.settings['config'])
        return SpawnPermissionsController.get_instance(cdsconfig, self.db).can_user_spawn(user.orm_user)



"""Added templates

Revision ID: f353692d88d5
Revises: 260ac5c1a9e0
Create Date: 2021-06-17 18:06:09.065414

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f353692d88d5'
down_revision = '260ac5c1a9e0'
branch_labels = None
depends_on = None

naming_convention = {
    "fk":
    "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
}

def upgrade():
    with op.batch_alter_table('dashboards', schema=None, naming_convention=naming_convention) as batch_op:
        batch_op.drop_constraint("fk_dashboards_source_spawner_id_spawners", type_='foreignkey')
        batch_op.drop_column('source_spawner_id')
        batch_op.add_column(sa.Column('template_parent_id', sa.Integer(), nullable=True))
        batch_op.create_foreign_key("dashboards_template_fk", 'dashboards', ['template_parent_id'], ['id'], ondelete='SET NULL')


def downgrade():
    with op.batch_alter_table('dashboards', schema=None) as batch_op:
        batch_op.add_column(sa.Column('source_spawner_id', sa.INTEGER(), nullable=True))
        batch_op.create_foreign_key(None, 'spawners', ['source_spawner_id'], ['id'], ondelete='SET NULL')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_column('template_parent_id')



"""Base

Revision ID: 2478e08b1ba2
Revises: 
Create Date: 2020-05-24 10:20:18.384756

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2478e08b1ba2'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



import os
pjoin = os.path.join

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

__version__ = None
exec(open(pjoin(here,'cdsdashboards/version.py')).read()) # Load __version__

install_requires = [
    'tornado>=6.0.4', # Mainly for jhsingle-native-proxy but best to keep consistent
    'traitlets',
    'jupyterhub>=1.0.0',
    'alembic',
    'pluggy'
]

extras_require = {
    'user': [
        'jhsingle-native-proxy>=0.7.6',
        'plotlydash-tornado-cmd>=0.0.6',
        'bokeh-root-cmd>=0.1.2',
        'rshiny-server-cmd>=0.0.2'
    ],
    'sudospawner': [
        "sudospawner>=0.5.2"
    ]
}

with open(pjoin(here, 'README.md'), encoding="utf8") as f:
    readme = f.read().replace('./docs/_static/screenshots/', 'https://github.com/ideonate/cdsdashboards/raw/master/docs/_static/screenshots/')

setup_metadata=dict(
    version=__version__,
    python_requires='>=3.6',
    author='Ideonate',
    author_email='dan@containds.com',
    license='BSD',
    url='https://github.com/ideonate/cdsdashboards',
    # this should be a whitespace separated string of keywords, not a list
    keywords="containds jupyterhub",
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    project_urls={
        'Source': 'https://github.com/ideonate/cdsdashboards',
        'Tracker': 'https://github.com/ideonate/cdsdashboards/issues'
    },
    long_description=readme,
    long_description_content_type='text/markdown',
    platforms="Linux, Mac OS X",
    description="ContainDS Dashboards extension for JupyterHub"
    )

# Data files e.g. templates and static js

share_cdsdashboards = pjoin(here, 'share', 'cdsdashboards')

def get_data_files():
    """Get data files in share/cdsdashboards"""

    data_files = []
    ntrim = len(here + os.path.sep)

    for (d, _, filenames) in os.walk(share_cdsdashboards):
        data_files.append((d[ntrim:], [pjoin(d, f)[ntrim:] for f in filenames]))
    return data_files

def get_package_data():
    """Get package data

    (mostly alembic config)
    """
    package_data = {}
    package_data['cdsdashboards'] = ['alembic.ini', 'cdsalembic/*', 'cdsalembic/versions/*']
    return package_data


setup_metadata.update(dict(
    name='cdsdashboards',
    packages=find_packages(),
    package_data=get_package_data(),
    data_files=get_data_files(),
    install_requires=install_requires,
    extras_require=extras_require
))

setup_metadata.update(dict(
    entry_points={
        'console_scripts': ['cds_sudospawner=cdsdashboards.hubextension.spawners.variablesudospawner:mediator']
    }))

setup(
    **setup_metadata
)



from traitlets import default
from kubespawner import KubeSpawner

from .variablemixin import VariableMixin, MetaVariableMixin, Command


class VariableKubeSpawner(KubeSpawner, VariableMixin, metaclass=MetaVariableMixin):

    default_presentation_cmd = Command(
        ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main'],
        allow_none=False,
        help="""
        The command to run presentations through jhsingle_native_proxy, can be a string or list.
        Default is ['python3', '-m', 'jhsingle_native_proxy.main']
        Change to e.g. ['start.sh', 'python3', '-m', 'jhsingle_native_proxy.main'] to ensure start hooks are
        run in the singleuser Docker images.
        """
    ).tag(config=True)
    
    @default("options_from_form")
    def _options_from_form_default(self):
        """
        Make sure we wrap the default options_from_form in a function that will preserve dashboard metadata
        May need to override this to something more specific, e.g. in KubeSpawner
        """
        usefn = self._default_options_from_form  # Spawner class default, just in case
        if hasattr(self, "_options_from_form"):  # KubeSpawner >= 1.0
            usefn = self._options_from_form
        elif hasattr(self, "options_from_form"):  # KubeSpawner < 1.0, although options_from_form fn will override this default anyway
            usefn = self.options_from_form

        return self._wrap_options_from_form(usefn)

